
#include "image/Image.hpp"

#include <stdint.h>

namespace jafar {
namespace image {

namespace details {
  
template<unsigned int depth, unsigned int sobeldepth, typename worktype>
double orientation(const Image* src_, double sigma = 1.5, bool usehistogram = false)
{
  Image imgx(src_->width(), src_->height(), sobeldepth, JfrImage_CS_GRAY);
  Image imgy(src_->width(), src_->height(), sobeldepth, JfrImage_CS_GRAY );
  cv::Sobel(*src_, imgx, sobeldepth, 1, 0, 3);
  cv::Sobel(*src_, imgy, sobeldepth, 0, 1, 3);

  int width = src_->width();
  int height = src_->height();
  double a = 1. / (8. * sigma * sigma); // 8 = 2 * 4 (2 pour la gaussian
  double sumh = 0.;
  double sumv = 0.;
  for(int j = 1; j < height-1; j++)
  {
    worktype* itx = reinterpret_cast<worktype*>(imgx.data(j)) + 1;
    worktype* ity = reinterpret_cast<worktype*>(imgy.data(j)) + 1;
    for(int i = 1; i < width-1; i++, itx++, ity++)
    {
      double h = *itx;
      double v = *ity;
      int x = (i - width/2);
      int y = (j - height/2);
      h *= exp( a * ( x*x  + y*y ) );
      v *= exp( a * ( x*x  + y*y ) );
      sumh +=h;
      sumv +=v;
    }
  }
  return atan2(sumv,sumh);
}
}

double Image::orientation(double sigma)
{
  switch(this->depth())
  {
    case CV_8U:
      return details::orientation<CV_8U, CV_16S, int16_t >(this,sigma);
    case CV_16U:
      return details::orientation<CV_16U, CV_16S, int16_t >(this,sigma);
    case CV_32F:
      return details::orientation<CV_32F, CV_32F, float>(this,sigma);
    case CV_64F:
      return details::orientation<CV_64F, CV_64F, double>(this,sigma);
    default:
      JFR_PRECOND(false, "Unknown image depth");
      return FP_NAN;
  }
}

}

}
