
#include "image/Image.hpp"
#include "boost/timer.hpp"

using namespace jafar::image;


template<typename worktype>
void applyNonOptimRankTransform (Image const* rawIma, const int radius, Image* filteredIma)
{  
  
  // pointers on image data
  worktype const* rImaPtr = reinterpret_cast<worktype const*>(rawIma->data());
  unsigned char* fImaPtr = filteredIma->data();
  
  // Image dimension
  unsigned int width = rawIma->width();
  unsigned int height = rawIma->height();
  unsigned int filStep = filteredIma->step();

  // Variables for rank Transform
  unsigned int R = radius, N = (radius*2+1);
  unsigned int testNumber = N*N-1;
  
  long int *shiftsTabular;
 
  // Memory Allocation
  shiftsTabular = new long int [testNumber];
  
  //
  // STEP #1 - Some pre-caclulations
  // 
 
  // bitIndexFilling filling
  unsigned int icol = 0, irow = 0, pos = 0;

  for (irow=0, pos=0;irow<N;++irow)
    for (icol=0;icol<N;++icol,++pos)
      {
	if ((icol==R) && (irow==R))
	  {
	    --pos;
	    continue;
	  }
	shiftsTabular[pos] = (irow-R)*filStep+icol-R;
      }
  
  //
  // STEP #2 - Rank Transform
  //   
  
  for (pos=0;pos<filStep*height;++pos)
    fImaPtr[pos] = 0;
  
  unsigned int current_pos, tested_pos;
  worktype current_gl;

  for (pos=0;pos<filStep*height;++pos)
    fImaPtr[pos] = 0;
    
  for (irow = R; irow<(height-R) ; ++irow)
    {
      pos = irow*filStep;
      for (icol = R; icol<(width-R) ; ++icol)
	{
	  current_pos = pos + icol;
	  current_gl = rImaPtr[current_pos];
	  for (unsigned int cpt_test=0;cpt_test<testNumber;++cpt_test)
	    {
	      tested_pos = current_pos+shiftsTabular[cpt_test];
	      if (current_gl<rImaPtr[tested_pos])
		fImaPtr[current_pos]++; 
	    }
	}
    }
  

  // Export image Block1
  filteredIma->save("/home/msanfour/versionnonoptimisee.tiff","vnonoptim");

  // Free memory
  delete[] shiftsTabular;

  
}



template<typename worktype>
void applyRankTransform (Image const* rawIma, const int radius, Image* filteredIma)
{  
  
  // pointers on image data
  worktype const* rImaPtr = reinterpret_cast<worktype const*>(rawIma->data());
  unsigned char* fImaPtr = filteredIma->data();
  
  // Image dimension
  unsigned int width = rawIma->width();
  unsigned int height = rawIma->height();
  //unsigned int rawStep = rawIma->step();
  unsigned int filStep = filteredIma->step();

  // Variables for rank Transform
  unsigned int R = radius, N = (radius*2+1);
  unsigned int maxMissingTestNumber = R*(R+1);
  unsigned int effectiveMTN;
  unsigned int testNumber = (N*N-1)/2;

  unsigned int *bitIndexTabular, *missingBitIndexTabular;
  long int *shiftsTabular, *missingShiftsTabular;
 
  // Memory Allocation
  bitIndexTabular = new unsigned int [N*N];
  shiftsTabular = new long int [testNumber];
  missingBitIndexTabular = new unsigned int [maxMissingTestNumber];
  missingShiftsTabular = new long int [maxMissingTestNumber];
  
  //
  // STEP #1 - Some pre-caclulations
  // 
 
  // bitIndexFilling filling
  unsigned int R_int=0, perim = 0, halfperim = 0, firstindex = 0, index = 0;
  unsigned int icol = 0, irow = 0, pos = 0;

  for (unsigned int iorbit=0;iorbit<R;iorbit++)
    {
      R_int = R-iorbit;
      perim = 8*R_int;
      halfperim = 4*R_int;
      
      index = firstindex;

      for (icol=iorbit;icol<(N-iorbit);++icol)
	{
	  bitIndexTabular[iorbit*N+icol] = index;
	  bitIndexTabular[(2*R-iorbit)*N+2*R-icol] = index + halfperim;
	  index++;
	}      
      --icol;

      for (irow=iorbit+1;irow<(N-iorbit-1);irow++)
	{
	  bitIndexTabular[irow*N+icol] = index;
	  bitIndexTabular[(2*R-irow)*N+2*R-icol] = index + halfperim;
	  index++;
	}

      firstindex += perim;
      
    }

  bitIndexTabular[testNumber] = 0;
  
  // shiftsTabularFilling
  for (irow = 0; irow<R ; ++irow)
    for (icol = 0; icol<N; ++icol, ++pos)
      shiftsTabular[pos] = (irow-R)*filStep+(icol-R);
  
  for (icol = 0; icol<R; ++icol, ++pos)
    shiftsTabular[pos] = icol-R;
 
  
  //
  // STEP #2 - Rank Transform
  //   
  
  unsigned int current_pos, tested_pos;
  worktype current_gl;

  for (pos=0;pos<filStep*height;++pos)
    fImaPtr[pos] = 0;
    
  // 1st Block
  for (irow = R; irow<height ; ++irow)
    {
      pos = irow*filStep;
      for ( icol = R; icol<(width-R) ; ++icol)
	{
	  current_pos = pos+icol;
	  current_gl = rImaPtr[current_pos];
	  for (unsigned int cpt_test=0;cpt_test<testNumber;++cpt_test)
	    {
	      tested_pos = current_pos+shiftsTabular[cpt_test];
	      if (current_gl!=rImaPtr[tested_pos])
        {
          if (current_gl<rImaPtr[tested_pos])
            fImaPtr[current_pos]++;
          else
            fImaPtr[tested_pos]++;
        }
      }
      }
    }

  // R-first effective columns 
  for (icol=R;icol<2*R;icol++)
    {      
      // Find missing tests
      effectiveMTN = R*(2*R-icol);
      for (unsigned int ccol = 0, pos = 0; ccol<(2*R-icol); ++ccol)
	for(irow = 1; irow<=R;++irow,++pos)
	  {
	    missingShiftsTabular[pos] = irow*filStep-R+ccol;
	    missingBitIndexTabular[pos] = bitIndexTabular[(irow+R)*N+ccol];
	  }
      
      // Make tests
      for (irow=R;irow<(height-R);++irow)
 	{
	  current_pos = irow*filStep+icol;
	  current_gl = rImaPtr[current_pos];
	  for (unsigned int cpt_test=0;cpt_test<effectiveMTN;++cpt_test)
	    {
	      tested_pos = current_pos+missingShiftsTabular[cpt_test];
	      if (current_gl<rImaPtr[tested_pos])
		fImaPtr[current_pos]++;
	    }
 	}
    }

  // R-last effective colums
  
  for (icol=(width-2*R);icol<(width-R);++icol)
    {
      // Find missing tests
      effectiveMTN = (R+1)*(icol-width+2*R+1);

      for (unsigned int ccol = R, pos = 0; ccol>(R-icol+width-2*R-1); --ccol)
	for(irow = 0; irow<=R;++irow,++pos)
	  {
 	    missingShiftsTabular[pos] = irow*filStep+ccol;
 	    missingBitIndexTabular[pos] = bitIndexTabular[(irow+R)*N+ccol+R];
	  }
      
      // Make tests
      for (irow=R;irow<(height-R);++irow)
  	{
 	  current_pos = irow*filStep+icol;
 	  current_gl = rImaPtr[current_pos];
 	  for (unsigned int cpt_test=0;cpt_test<effectiveMTN;++cpt_test)
 	    {
 	      tested_pos = current_pos+missingShiftsTabular[cpt_test];
 	      if (current_gl<rImaPtr[tested_pos])
 		fImaPtr[current_pos]++;
 	    }
  	}
    }
      
  // Export image Block1
  filteredIma->save("/home/msanfour/versionoptimisee.tiff","voptim");

  // Free memory
  delete[] bitIndexTabular;
  delete[] shiftsTabular;
  delete[] missingBitIndexTabular;
  delete[] missingShiftsTabular;
  
}


Image* Image::rankTransform (const int radius)
{
  JFR_PRECOND(radius<=7, "rankTransform : radius must be inferior or equal to 7 ");
  JFR_PRECOND(this->colorSpace()==JfrImage_CS_GRAY, "rankTransform works only on greyscaled images");
  
  Image* filteredImage = new Image(this->width(), this->height(), CV_8U, this->colorSpace());

  boost::timer t1;
  switch(this->depth())
    {
//    case CV_1U:
//      {
//	applyRankTransform<bool>(this,radius,filteredImage);
//	break;
//      }
    case CV_8U:
      {
	applyRankTransform<uint8_t>(this,radius,filteredImage);
	break;
      }
    case CV_8S:
      {
	applyRankTransform<int8_t>(this,radius,filteredImage);
	break;
      }
    case CV_16U:
      {
	applyRankTransform<uint16_t>(this,radius,filteredImage);
	break;
      }
    case CV_16S:
      {
	applyRankTransform<int16_t>(this,radius,filteredImage);
	break;
      }
    case CV_32F:
      {
	applyRankTransform<float>(this,radius,filteredImage);
	break;
      }
    }
  std::cout << " Optimized version : " << t1.elapsed() << std::endl;
  return filteredImage;
  
}
